@@ -3,24 +3,27 @@
 #' @description
 #' `scale_and_register_data()` is a function to register expression profiles a user wish to compare. This includes an option to scale data before registration, find and calculate score of optimal shifts and stretches, as well as apply the best shifts and stretches.
 #'
-#' @param input_df Input data frame contains all replicates of gene expression in each genotype at each time point.
+#' @param input_df Input data frame containing all replicates of gene expression in each genotype at each time point.
 #' @param stretches Candidate registration stretch factors to apply to data to transform.
-#' @param shift_extreme The absolute maximum value which can be applied as a shift to gene expression time course (days).
-#' @param num_shifts Number of shifts between minimum and maximum values of shift.
-#' @param min_num_overlapping_points Number of minimum overlapping time points.  Shifts will be only considered if it leaves at least these many overlapping points after applying the registration function.
+#' @param shifts Candidate registration shift values to apply to data to transform.
+#' @param min_num_overlapping_points Number of minimum overlapping time points. Shifts will be only considered if it leaves at least these many overlapping points after applying the registration function.
+#' @param maintain_min_num_overlapping_points Whether to automatically calculate extreme (minimum and maximum) values of \code{shifts} to maintain specified \code{min_num_overlapping_points} condition. By default, \code{FALSE}.
 #' @param initial_rescale Scaling gene expression prior to registration if \code{TRUE}.
 #' @param do_rescale Scaling gene expression using only overlapping time points points during registration.
 #' @param accession_data_to_transform Accession name of data which will be transformed.
 #' @param accession_data_ref Accession name of reference data.
 #' @param start_timepoint Time points to be added in both reference data and data to transform after shifting and stretching. Can be either \code{"reference"} (the default), \code{"transform"}, or \code{"zero"}.
 #' @param expression_value_threshold Expression value threshold. Remove expressions if maximum is less than the threshold. If \code{NULL} keep all data.
-#' @param is_data_normalised TRUE if dataset has been normalised prior to registration process.
+#' @param is_data_normalised \code{TRUE} if dataset has been normalised prior to registration process.
+#' @param optimise_registration_parameters Whether to optimise registration parameters with Simulated Annealing. By default, \code{FALSE}.
+#' @param num_iterations Maximum number of iterations in the Simulated Annealing optimisation. By default, 60.
 #'
 #' @return This function returns a list of data frames, containing:
 #'
 #' \item{mean_df}{a data frame containing mean expression value of each gene and accession for every time point.}
 #' \item{mean_df_sc}{identical to `mean_df`, with additional column `sc.expression_value` which the scaled mean expression values.}
 #' \item{to_shift_df}{a processed input data frame which is ready to be registered.}
+#' \item{best_shifts}{a data frame containing best shift factor for each given stretch.}
 #' \item{shifted_mean_df}{the registration result - after stretching and shifting.}
 #' \item{imputed_mean_df}{the imputed (transformed to be the same in a set of common time points) registration result.}
 #' \item{all_shifts_df}{a table containing candidates of registration parameters and a score after applying each parameter (stretch and shift factor).}
@@ -38,8 +41,7 @@
 #' registration_results <- scale_and_register_data(
 #'   input_df = all_data_df,
 #'   stretches = c(3, 2.5, 2, 1.5, 1),
-#'   shift_extreme = 4,
-#'   num_shifts = 27,
+#'   shifts = seq(-4, 4, length.out = 33),
 #'   min_num_overlapping_points = 4,
 #'   initial_rescale = FALSE,
 #'   do_rescale = TRUE,
@@ -49,109 +51,125 @@
 #' )
 #' }
 scale_and_register_data <- function(input_df,
-                                    stretches,
-                                    shift_extreme,
-                                    num_shifts,
+                                    stretches = NA,
+                                    shifts = NA,
                                     min_num_overlapping_points,
-                                    initial_rescale,
-                                    do_rescale,
+                                    maintain_min_num_overlapping_points = FALSE,
+                                    initial_rescale = FALSE,
+                                    do_rescale = TRUE,
                                     accession_data_to_transform,
                                     accession_data_ref,
                                     start_timepoint = c("reference", "transform", "zero"),
                                     expression_value_threshold = 5,
-                                    is_data_normalised = FALSE) {
+                                    is_data_normalised = FALSE,
+                                    optimise_registration_parameters = FALSE,
+                                    num_iterations = 60) {
   # Validate parameters
   start_timepoint <- match.arg(start_timepoint)
 
-  # Suppress "no visible binding for global variable" note
-  accession <- NULL
-  timepoint <- NULL
-  sc.expression_value <- NULL
-  expression_value <- NULL
-  locus_name <- NULL
-
-  # Make sure the data are data.tables
-  all_data_df <- data.table::as.data.table(input_df)
-
-  mean_df <- get_mean_data(
-    exp = all_data_df,
-    expression_value_threshold = expression_value_threshold,
+  # Preprocess data
+  processed_data <- preprocess_data(
+    input_df = input_df,
+    initial_rescale = initial_rescale,
     accession_data_to_transform = accession_data_to_transform,
+    accession_data_ref = accession_data_ref,
+    start_timepoint = start_timepoint,
+    expression_value_threshold = expression_value_threshold,
     is_data_normalised = is_data_normalised
   )
 
-  # Parse start_timepoint
-  if (start_timepoint == "reference") {
-    time_to_add <- min(all_data_df[accession == accession_data_ref, timepoint])
-  } else if (start_timepoint == "transform") {
-    time_to_add <- min(all_data_df[accession == accession_data_to_transform, timepoint])
+  all_data_df <- processed_data$all_data_df
+  mean_df <- processed_data$mean_df
+  mean_df_sc <- processed_data$mean_df_sc
+  to_shift_df <- processed_data$to_shift_df
+  time_to_add <- processed_data$time_to_add
+
+  if (!optimise_registration_parameters) {
+    # Check that stretches and shifts are numeric
+    if (any(is.na(stretches)) | any(is.na(shifts))) {
+      stop(cli::format_error(c(
+        "{.var stretches} and {.var shifts} must be numeric vectors",
+        "x" = "You supplied vectors with {.cls NA} values."
+      )))
+    }
+
+    # Proceed with original registration pipeline
+    cli::cli_h1("Information before registration")
+    cli::cli_alert_info("Max value of expression_value of all_data_df: {cli::col_cyan(round(max(all_data_df$expression_value), 2))}")
+
+    # Calculate the best registration. Returns all tried registrations, best stretch and shift combo, and BIC stats for comparison of best registration model to separate models for expression of each gene in reference data and data to transform
+    cli::cli_h1("Analysing models for all stretch and shift factor")
+    best_registration_list <- get_best_stretch_and_shift(
+      to_shift_df,
+      all_data_df,
+      stretches,
+      shifts,
+      do_rescale,
+      min_num_overlapping_points,
+      maintain_min_num_overlapping_points,
+      accession_data_to_transform,
+      accession_data_ref,
+      time_to_add
+    )
+
+    all_shifts <- best_registration_list$all_shifts
+    best_shifts <- best_registration_list$best_shifts
+    model_comparison_dt <- best_registration_list$model_comparison_dt
   } else {
-    time_to_add <- 0
+    # Check that stretches and shifts are both numeric or both NA
+    if (length(unique(c(any(is.na(stretches)), any(is.na(shifts))))) == 2) {
+      stop(cli::format_error(c(
+        "{.var stretches} and {.var shifts} must either both be defined or both be {.cls NA}.",
+        "x" = "You supplied {.var stretches} as {.cls {ifelse(any(is.na(stretches)), 'NA', class(stretches))}} and {.var shifts} as {.cls {ifelse(any(is.na(shifts)), 'NA', class(shifts))}}."
+      )))
+    }
+
+    # Optimise registration parameters
+    cli::cli_h1("Starting optimisation")
+
+    optimised_parameters <- optimise_registration_params(
+      input_df = input_df,
+      genes = NULL,
+      stretches_bound = stretches,
+      shifts_bound = shifts,
+      initial_rescale = initial_rescale,
+      do_rescale = do_rescale,
+      min_num_overlapping_points = min_num_overlapping_points,
+      maintain_min_num_overlapping_points = maintain_min_num_overlapping_points,
+      accession_data_to_transform = accession_data_to_transform,
+      accession_data_ref = accession_data_ref,
+      start_timepoint = start_timepoint,
+      expression_value_threshold = expression_value_threshold,
+      is_data_normalised = is_data_normalised,
+      num_iterations = num_iterations
+    )
+
+    cli::cli_alert_success("Finished optimisation")
+
+    # Mimic original pipeline data structure to proceed with apply_shift_to_registered_genes_only()
+    best_registration_list <- get_best_stretch_and_shift_after_optimisation(
+      to_shift_df,
+      all_data_df,
+      optimised_parameters,
+      do_rescale,
+      min_num_overlapping_points,
+      accession_data_to_transform,
+      accession_data_ref,
+      time_to_add
+    ) %>%
+      suppressMessages()
+
+    all_shifts <- best_registration_list$all_shifts
+    best_shifts <- best_registration_list$best_shifts
+    model_comparison_dt <- best_registration_list$model_comparison_dt
   }
 
-  # Filter genes of original input data as in mean_df
-  all_data_df <- all_data_df[all_data_df$locus_name %in% unique(mean_df$locus_name)]
-  all_data_df <- subset(all_data_df, select = c("locus_name", "accession", "tissue", "timepoint", "expression_value", "group"))
-
-  # TODO: validate colnames
-  # Apply normalisation of expression for each gene across all timepoints
-  mean_df_sc <- data.table::copy(mean_df)
-
-  # Apply scaling
-  mean_df_sc[, sc.expression_value := scale(expression_value, scale = TRUE, center = TRUE), by = .(locus_name, accession)]
-
-  # Apply scaling before registration (if initial_rescale == TRUE), otherwise using original data
-  if (initial_rescale == TRUE) {
-
-    # apply rescale to mean_df prior to registration
-    to_shift_df <- data.table::copy(mean_df_sc)
-    to_shift_df$expression_value <- to_shift_df$sc.expression_value
-    to_shift_df$sc.expression_value <- NULL
-
-    # apply THE SAME rescale to all_data_df prior to registration
-    all_data_df <- scale_all_rep_data(mean_df, all_data_df, "scale")
-  } else {
-    to_shift_df <- data.table::copy(mean_df)
-  }
-
-  cli::cli_h1("Information before registration")
-  cli::cli_alert_info("Max value of expression_value of all_data_df: {cli::col_cyan(round(max(all_data_df$expression_value), 2))}")
-
-  # Calculate the best registration. Returns all tried registrations, best stretch and shift combo,and AIC/BIC stats for comparison of best registration model to separate models for expression ofeach gene in Ro18 and Col0
-  cli::cli_h1("Analysing models for all stretch and shift factor")
-
-  L <- get_best_stretch_and_shift(
-    to_shift_df,
-    all_data_df,
-    stretches,
-    do_rescale,
-    min_num_overlapping_points,
-    shift_extreme,
-    num_shifts,
-    accession_data_to_transform,
-    accession_data_ref,
-    time_to_add
-  )
-
-  all_shifts <- L[["all_shifts"]]
-  best_shifts <- L[["best_shifts"]]
-  model_comparison_dt <- L[["model_comparison_dt"]]
-
-  # Add columns which flags which BIC and AIC values are better
-  model_comparison_dt$BIC_registered_is_better <- (model_comparison_dt$registered.BIC < model_comparison_dt$separate.BIC)
-  # model_comparison_dt$AIC_registered_is_better <- (model_comparison_dt$registered.AIC < model_comparison_dt$separate.AIC)
-  # model_comparison_dt$ABIC_registered_is_better <- (model_comparison_dt$BIC_registered_is_better & model_comparison_dt$AIC_registered_is_better)
-
-
   # Report model comparison results
   cli::cli_h1("Model comparison results")
-  # cli::cli_alert_info("AIC finds registration better than non-registration for: {cli::col_cyan(sum(model_comparison_dt$AIC_registered_is_better), '/', nrow(model_comparison_dt))}")
   cli::cli_alert_info("BIC finds registration better than non-registration for: {cli::col_cyan(sum(model_comparison_dt$BIC_registered_is_better), '/', nrow(model_comparison_dt))}")
-  # cli::cli_alert_info("AIC & BIC finds registration better than non-registration for: {cli::col_cyan(sum(model_comparison_dt$ABIC_registered_is_better), '/', nrow(model_comparison_dt))}")
-
-  # Get the best-shifted and stretched mean gene expression, only to genes which registration is better than
-  # separate models by BIC. Don't stretch out, or shift genes for which separate is better.
 
+  # Get the best-shifted and stretched mean gene expression, only to genes which registration is better
+  # than separate models by BIC. Don't stretch out, or shift genes for which separate is better.
   cli::cli_h1("Applying the best-shifts and stretches to gene expression")
   shifted_mean_df <- apply_shift_to_registered_genes_only(
     to_shift_df,
@@ -164,7 +182,6 @@ scale_and_register_data <- function(input_df,
 
   cli::cli_alert_info("Max value of expression_value: {cli::col_cyan(round(max(shifted_mean_df$expression_value), 2))}")
 
-
   # Impute transformed values at times == to the observed reference data points for each shifted transformed gene so can compare using heat maps.
   # Transformed curves are the ones that been shifted around. Linear impute values for these curves so that reference data samples can be compared to an transformed data point.
   imputed_mean_df <- impute_transformed_exp_values(
@@ -173,15 +190,19 @@ scale_and_register_data <- function(input_df,
     accession_data_ref
   )
 
-  out <- list(
+  # Results object
+  results_list <- list(
     "mean_df" = mean_df,
     "mean_df_sc" = mean_df_sc,
     "to_shift_df" = to_shift_df,
+    "best_shifts" = best_shifts,
     "shifted_mean_df" = shifted_mean_df,
     "imputed_mean_df" = imputed_mean_df,
     "all_shifts_df" = all_shifts,
     "model_comparison_df" = model_comparison_dt
   )
+
+  return(results_list)
 }
 
 #' Scaling all un-averaged data
@@ -238,16 +259,16 @@ scale_all_rep_data <- function(mean_df, all_rep_data, scale_func) {
   return(out)
 }
 
-#' Calculate best shifts and stretches for each gene, also calculate AIC/BIC under registration or non-registration
+#' Calculate best shifts and stretches for each gene, also calculate BIC under registration or non-registration
 #'
 #' @noRd
 get_best_stretch_and_shift <- function(to_shift_df,
                                        all_data_df,
                                        stretches,
+                                       shifts,
                                        do_rescale,
                                        min_num_overlapping_points,
-                                       shift_extreme,
-                                       num_shifts,
+                                       maintain_min_num_overlapping_points,
                                        accession_data_to_transform,
                                        accession_data_ref,
                                        time_to_add) {
@@ -271,19 +292,20 @@ get_best_stretch_and_shift <- function(to_shift_df,
 
     # Calculate all the shift scores given this stretch. Score is mean(dist^2), over overlapping points if do_rescale=T, is rescaled by the mean FOR THE OVERLAPPING POINTS. (but not by the SD.)
     all_shifts <- calculate_all_best_shifts(
-      num_shifts,
       mean_df = to_shift_df,
       stretch_factor = stretch,
+      shifts,
       do_rescale,
-      shift_extreme,
       min_num_overlapping_points,
+      maintain_min_num_overlapping_points,
       accession_data_to_transform,
       accession_data_ref
     )
 
-    all_shifts <- unique(all_shifts) # ensure no duplicated rows
+    # Ensure no duplicated rows
+    all_shifts <- unique(all_shifts)
 
-    # Cut down to single best shift for each gene
+    # Cut down to single best shift for each gene (Alex's original logic)
     all_shifts[, is_best := get_best_result(.SD), by = .(gene)]
     best_shifts <- all_shifts[is_best == TRUE, ]
     all_shifts$is_best <- NULL
@@ -292,7 +314,7 @@ get_best_stretch_and_shift <- function(to_shift_df,
       stop("get_best_stretch_and_shift(): got non-unique best shifts in best_shifts")
     }
 
-    # Calculate the BIC & AIC for the best shifts found with this stretch.compared to treating the gene's expression separately in data to transform and reference data
+    # Calculate the BIC for the best shifts found with this stretch.compared to treating the gene's expression separately in data to transform and reference data
     model_comparison_dt <- calculate_all_model_comparison_stats(
       all_data_df,
       best_shifts,
@@ -304,7 +326,7 @@ get_best_stretch_and_shift <- function(to_shift_df,
     # Add info on the stretch and shift applied
     model_comparison_dt <- merge(
       model_comparison_dt,
-      best_shifts[, c("gene", "stretch", "shift"), ],
+      best_shifts[, c("gene", "stretch", "shift", "score"), ],
       by = "gene"
     )
 
@@ -322,12 +344,12 @@ get_best_stretch_and_shift <- function(to_shift_df,
   # model comparison of best shift (for each stretch) to separate models
   all_model_comparison_dt <- do.call("rbind", all_model_comparison_dt)
 
-  # Correct -Inf BIC and AIC values to -9999 so that delta.BIC is not Inf or NaN
+  # Correct -Inf BIC values to -9999 so that delta.BIC is not Inf or NaN
   all_model_comparison_dt <- all_model_comparison_dt %>%
     dplyr::rowwise() %>%
     dplyr::mutate(
       dplyr::across(
-        .cols = c(.data$registered.BIC, .data$registered.AIC, .data$separate.BIC, .data$separate.AIC),
+        .cols = c(.data$registered.BIC, .data$separate.BIC),
         .fns = function(x) {
           if (!is.finite(x)) {
             x <- 9999 * sign(x)
@@ -353,6 +375,7 @@ get_best_stretch_and_shift <- function(to_shift_df,
     best_model_comparison.dt <- best_model_comparison.dt[!(duplicated(best_model_comparison.dt$gene)), ]
   }
 
+  best_model_comparison.dt$BIC_registered_is_better <- best_model_comparison.dt$delta.BIC < 0
   best_model_comparison.dt$delta.BIC <- NULL
 
   # Cut down best shifts to the best shift for the best stretch only
@@ -367,11 +390,14 @@ get_best_stretch_and_shift <- function(to_shift_df,
     stop()
   }
 
-  return(list(
+  # Results object
+  results_list <- list(
     "all_shifts" = all_shifts,
     "best_shifts" = best_shifts,
     "model_comparison_dt" = best_model_comparison.dt
-  ))
+  )
+
+  return(results_list)
 }
 
 #' Apply shift for all registered genes
@@ -443,7 +469,6 @@ impute_transformed_exp_values <- function(shifted_mean_df,
   # The imputed transformed data times going to estimate gene expression for
   imputed_timepoints <- round(seq(min(shifted_mean_df$shifted_time), max(shifted_mean_df$shifted_time)))
 
-  # browser()
   out_list <- list()
   out_list <- c(out_list, list(shifted_mean_df[shifted_mean_df$accession == accession_data_ref]))
 
@@ -480,7 +505,81 @@ impute_transformed_exp_values <- function(shifted_mean_df,
     i <- i + 1
   }
 
+  # Bind results
   out_df <- do.call("rbind", out_list)
 
   return(out_df)
 }
+
+
+#' Preprocess data before registration
+#'
+#' @noRd
+preprocess_data <- function(input_df,
+                            initial_rescale,
+                            accession_data_to_transform,
+                            accession_data_ref,
+                            start_timepoint,
+                            expression_value_threshold,
+                            is_data_normalised) {
+  # Suppress "no visible binding for global variable" note
+  accession <- NULL
+  timepoint <- NULL
+  sc.expression_value <- NULL
+  expression_value <- NULL
+  locus_name <- NULL
+
+  # Make sure the data are data.tables
+  all_data_df <- data.table::as.data.table(input_df)
+
+  mean_df <- get_mean_data(
+    exp = all_data_df,
+    expression_value_threshold = expression_value_threshold,
+    accession_data_to_transform = accession_data_to_transform,
+    is_data_normalised = is_data_normalised
+  )
+
+  # Parse start_timepoint
+  if (start_timepoint == "reference") {
+    time_to_add <- min(all_data_df[accession == accession_data_ref, timepoint])
+  } else if (start_timepoint == "transform") {
+    time_to_add <- min(all_data_df[accession == accession_data_to_transform, timepoint])
+  } else {
+    time_to_add <- 0
+  }
+
+  # Filter genes of original input data as in mean_df
+  all_data_df <- all_data_df[all_data_df$locus_name %in% unique(mean_df$locus_name)]
+  all_data_df <- subset(all_data_df, select = c("locus_name", "accession", "tissue", "timepoint", "expression_value", "group"))
+
+  # TODO: validate colnames
+  # Apply normalisation of expression for each gene across all timepoints
+  mean_df_sc <- data.table::copy(mean_df)
+
+  # Apply scaling
+  mean_df_sc[, sc.expression_value := scale(expression_value, scale = TRUE, center = TRUE), by = .(locus_name, accession)]
+
+  # Apply scaling before registration (if initial_rescale == TRUE), otherwise using original data
+  if (initial_rescale == TRUE) {
+    # Apply rescale to mean_df prior to registration
+    to_shift_df <- data.table::copy(mean_df_sc)
+    to_shift_df$expression_value <- to_shift_df$sc.expression_value
+    to_shift_df$sc.expression_value <- NULL
+
+    # Apply THE SAME rescale to all_data_df prior to registration
+    all_data_df <- scale_all_rep_data(mean_df, all_data_df, "scale")
+  } else {
+    to_shift_df <- data.table::copy(mean_df)
+  }
+
+  # Results object
+  results_list <- list(
+    all_data_df = all_data_df,
+    mean_df = mean_df,
+    mean_df_sc = mean_df_sc,
+    to_shift_df = to_shift_df,
+    time_to_add = time_to_add
+  )
+
+  return(results_list)
+}
